/* SPDX-License-Identifier: LGPL-2.1-or-later */
#pragma once

int verb_edit(int argc, char *argv[], void *userdata);
int verb_cat(int argc, char *argv[], void *userdata);

int verb_mask(int argc, char *argv[], void *userdata);
int verb_unmask(int argc, char *argv[], void *userdata);
